@@ -1,5 +1,7 @@
+//! This module contains abstractions of the libretro environment callbacks.
 use super::*;
 
+#[doc(hidden)]
 macro_rules! into_generic {
     ($type:ty, $lifetime:tt) => {
         into_generic!($type, GenericContext, $lifetime);
@@ -19,6 +21,7 @@ macro_rules! into_generic {
     };
 }
 
+#[doc(hidden)]
 macro_rules! make_context {
     ($name:ident $(, #[doc = $doc:tt ])?) => {
         $(#[doc = $doc])?
@@ -54,12 +57,14 @@ impl<'a> GenericContext<'a> {
         self.environment_callback
     }
 
+    /// Enables the [`Core::on_keyboard_event`] callback.
     pub fn enable_keyboard_callback(&self) -> bool {
         self.set_keyboard_callback(retro_keyboard_callback {
             callback: Some(retro_keyboard_callback_fn),
         })
     }
 
+    /// Enables the [`Core::on_write_audio`] and [`Core::on_audio_set_state`] callbacks.
     pub fn enable_audio_callback(&self) -> bool {
         self.set_audio_callback(retro_audio_callback {
             callback: Some(retro_audio_callback_fn),
@@ -114,26 +119,47 @@ impl<'a> GenericContext<'a> {
     }
 }
 
+/// Functions that are safe to be called in [`Core::on_reset`].
 pub type ResetContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_deinit`].
 pub type DeinitContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::get_serialize_size`].
 pub type GetSerializeSizeContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_serialize`].
 pub type SerializeContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_unserialize`].
 pub type UnserializeContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_unload_game`].
 pub type UnloadGameContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_cheat_reset`].
 pub type CheatResetContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_cheat_set`].
 pub type CheatSetContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::on_get_region`].
 pub type GetRegionContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::get_memory_data`].
 pub type GetMemoryDataContext<'a> = GenericContext<'a>;
+
+/// Functions that are safe to be called in [`Core::get_memory_size`].
 pub type GetMemorySizeContext<'a> = GenericContext<'a>;
 
-make_context!(GetAvInfoContext);
-make_context!(InitContext);
-make_context!(OptionsChangedContext);
+make_context!(GetAvInfoContext, #[doc = "Functions that are safe to be called in [`Core::on_get_av_info`]"]);
+make_context!(InitContext, #[doc = "Functions that are safe to be called in [`Core::on_init`]"]);
+make_context!(OptionsChangedContext, #[doc = "Functions that are safe to be called in [`Core::on_options_changed`]"]);
 
-make_context!(LoadGameSpecialContext);
+make_context!(LoadGameSpecialContext, #[doc = "Functions that are safe to be called in [`Core::on_load_game_special`]"]);
 // into_generic!(LoadGameSpecialContext<'a>, LoadGameContext, 'a);
 
-make_context!(SetEnvironmentContext);
+make_context!(SetEnvironmentContext, #[doc = "Functions that are safe to be called in [`Core::on_set_environment`]"]);
 
 impl<'a> SetEnvironmentContext<'a> {
     pub fn enable_proc_address_interface(&mut self) -> bool {
@@ -143,6 +169,9 @@ impl<'a> SetEnvironmentContext<'a> {
     }
 }
 
+/// Functions that are safe to be called in [`Core::on_load_game`].
+///
+/// For a description of the callbacks see [`CoreWrapper`].
 pub struct LoadGameContext<'a> {
     pub(crate) environment_callback: &'a retro_environment_t,
 
@@ -265,6 +294,9 @@ impl<'a> LoadGameContext<'a> {
 }
 into_generic!(LoadGameContext<'a>, 'a);
 
+/// Functions that are safe to be called in [`Core::on_write_audio`].
+///
+/// For a description of the callbacks see [`CoreWrapper`].
 pub struct AudioContext<'a> {
     pub(crate) environment_callback: &'a retro_environment_t,
 
@@ -273,7 +305,7 @@ pub struct AudioContext<'a> {
 }
 
 impl AudioContext<'_> {
-    /// Renders multiple audio frames in one go if [`CoreWrapper::audio_sample_batch_callback`] has been set.
+    /// Renders multiple audio frames in one go if [`AudioContext::audio_sample_batch_callback`] has been set.
     ///
     /// One frame is defined as a sample of left and right channels, interleaved.
     /// I.e. `let buf: [u16; 4] = [ l, r, l, r ];` would be 2 frames.
@@ -289,7 +321,7 @@ impl AudioContext<'_> {
         }
     }
 
-    /// Renders a single audio frame if [`CoreWrapper::audio_sample_callback`] has been set.
+    /// Renders a single audio frame if [`AudioContext::audio_sample_callback`] has been set.
     /// Should only be used if implementation generates a single sample at a time.
     /// Format is signed 16-bit native endian.
     ///
@@ -305,7 +337,7 @@ impl AudioContext<'_> {
 
 into_generic!(AudioContext<'a>, 'a);
 
-/// Exposes callbacks that are safe to call in [`Core::on_run`]
+/// Functions that are safe to be called in [`Core::on_run`].
 ///
 /// For a description of the callbacks see [`CoreWrapper`].
 pub struct RunContext<'a> {
@@ -344,7 +376,7 @@ impl<'a> RunContext<'_> {
         self.can_dupe
     }
 
-    /// Polls for input if [`CoreWrapper::input_poll_callback`] has been set
+    /// Polls for input if [`RunContext::input_poll_callback`] has been set
     pub fn poll_input(&self) {
         if let Some(callback) = self.input_poll_callback {
             unsafe {
@@ -353,7 +385,7 @@ impl<'a> RunContext<'_> {
         }
     }
 
-    /// Gets the input state for the given player and device if [`CoreWrapper::input_state_callback`] has been set
+    /// Gets the input state for the given player and device if [`RunContext::input_state_callback`] has been set
     pub fn get_input_state(&self, port: u32, device: u32, index: u32, id: u32) -> u16 {
         if let Some(callback) = self.input_state_callback {
             unsafe { (callback)(port, device, index, id) as u16 }
@@ -478,7 +510,7 @@ impl<'a> RunContext<'_> {
         JoypadState::empty()
     }
 
-    /// Draws a new frame if [`CoreWrapper::video_refresh_callback`] has been set
+    /// Draws a new frame if [`RunContext::video_refresh_callback`] has been set
     pub fn draw_frame(&mut self, data: &[u8], width: u32, height: u32, pitch: u64) {
         if let Some(callback) = self.video_refresh_callback {
             *self.had_frame = true;
